
class AppKeys:
    """
    Store application keys in one place.
    
    :var: GITHUB_TOKEN
    :var: VAULT_ADDR
    :var: IMAGE_API_RELATIVE_URL
    :var: IMAGE_ENHANCE_API_RELATIVE_URL
    """

    GITHUB_TOKEN = 'GITHUB_TOKEN'
    VAULT_ADDR = 'VAULT_ADDR'
    IMAGE_API_RELATIVE_URL = 'image_api_relative_url'
    IMAGE_ENHANCE_API_RELATIVE_URL = 'image_enhance_api_relative_url'


class ImageVariants:
    """
    Store allowed values for the image variant.
    
    :var: STANDARD
    :var: THUMBNAIL
    :var: ORIGINAL
    :var: DEFAULT
    """

    STANDARD = 'standard'
    THUMBNAIL = 'thumbnail'
    ORIGINAL = 'original'
    DEFAULT = STANDARD



from elvia_louvre.data_models import SecretPathSets, UpdateMetadataRequest
from elvia_louvre.token import Token
from io import BytesIO
from PIL import Image
import requests

from .config import Config
from .enums import ImageVariants
from .image_api import ImageAPI
from .image_enhance_api import ImageEnhanceAPI


class LouvreClient:
    """
    Python client to interact with APIs in the Louvre system.
    
    :param SecretPathSets secret_path_sets: SecretPathSets instance
    """

    _config: Config
    def __init__(self, secret_path_sets: SecretPathSets):
        self._config = Config(secret_path_sets=secret_path_sets)

    def get_image_data(self,
                       image_id: str,
                       using_production_images: bool = False):
        """
        Return an object with image metadata retrieved from ImageAPI.
        
        :param str image_id: ID of the image.
        :param using_production_images: Whether to search in production.
        :type using_production_images: bool, defaults to False

        :rtype: ImageData
        :raises errors.LouvreImageNotFound:
        :raises errors.LouvreKeyError:
        :raises errors.LouvreQueryError:
        :raises requests.RequestException:
        """
        return ImageAPI.get_image_data(
            config=self._config,
            img_id=image_id,
            using_production_images=using_production_images)

    def get_image_url(self,
                      image_id: str,
                      image_variant: str = ImageVariants.DEFAULT,
                      using_production_images: bool = False) -> str:
        """
        Return the URL of an image variant.

        :param str image_id:  ID of the image.
        :param str image_variant: Accepted values can be found in the ImageVariants class.
        :param using_production_images: Whether to search in production.
        :type using_production_images: bool, defaults to False        

        :rtype: str
        :raises errors.LouvreImageNotFound:
        :raises errors.LouvreInvalidImageVariant:
        :raises errors.LouvreKeyError:
        :raises errors.LouvreQueryError:
        :raises requests.RequestException:
        """
        return ImageAPI.get_image_sasuri(
            config=self._config,
            img_id=image_id,
            image_variant=image_variant,
            using_production_images=using_production_images)

    def get_image(self,
                  image_id: str,
                  image_variant: str = ImageVariants.DEFAULT,
                  using_production_images: bool = False) -> Image:
        """
        Return the image, as a PIL object.

        :param str image_id:  ID of the image.
        :param str image_variant: Accepted values can be found in the ImageVariants class.
        :param using_production_images: Whether to search in production.
        :type using_production_images: bool, defaults to False   

        :returns: PIL.Image
        :raises errors.LouvreImageNotFound:
        :raises errors.LouvreInvalidImageVariant:
        :raises errors.LouvreKeyError:
        :raises errors.LouvreQueryError:
        :raises requests.RequestException:
        """
        file_url = self.get_image_url(
            image_id=image_id,
            image_variant=image_variant,
            using_production_images=using_production_images)
        _response = requests.get(url=file_url)
        _file_bytes = BytesIO(_response.content)
        return Image.open(_file_bytes)

    def update_image_metadata(self,
                              update_request: UpdateMetadataRequest) -> None:
        """
        Update the metadata for an image.

        :param UpdateMetadataRequest update_request: Metadata update request instance
        :raises errors.LouvreQueryError:
        :raises requests.RequestException:
        """
        ImageEnhanceAPI.update_metadata(config=self._config,
                                        update_request=update_request)

    def get_headers(self, using_production_images: bool = False) -> dict:
        """
        Return request headers to be included in a call to Louvre APIs.

        :param using_production_images: Whether to search in production.
        :type using_production_images: bool, defaults to False    
        :rtype: dict
        :raises errors.LouvreSecretError:
        """
        return Token.get_headers(
            config=self._config,
            using_production_images=using_production_images)



from dataclasses import dataclass, field
from elvia_louvre.errors import LouvreSecretError
from typing import Dict, List, Optional, Union


@dataclass
class ImageFile:
    """
    Represent image files from ImageAPI.
    
    :param str image_variant: Name of the image variant for this file
    :param int size: File size in bytes.
    :param str sasuri: URL to the file. It contains a token with built-in expiring access to the resource.
    :param float height: Image height in pixels.
    :param float width: Image width in pixels.
    """

    image_variant: str
    size: int
    sasuri: str
    height: float
    width: float


@dataclass
class ImageData:
    """
    Represent image data from ImageAPI.
    
    :param str image_id:
    :param str etag:
    :param List[dict] metadata:
    :param List[data_models.ImageFile] image_files:
    """

    image_id: str
    etag: str
    metadata: List[dict]
    image_files: List[ImageFile]

    @property
    def image_variants(self) -> List[str]:
        """
        Return a list with the imageVariants present.

        :rtype: List[str]
        """
        return [image_file.image_variant for image_file in self.image_files]

    def get_variant(self, variant: str) -> Union[ImageFile, None]:
        """
        Given an image variant, return its corresponding ImageFile if it exits in the ImageData 
        instance.

        :param variant: Name of the desired image variant.
        :type variant: str            

        :rtype: data_models.ImageFile, optional
        """
        for image_file in self.image_files:
            if variant == image_file.image_variant:
                return image_file
        return None


@dataclass
class UpdateMetadataRequest:
    """
    Represent a metadata update request to be sent to ImageEnhanceAPI.
    
    :param str image_id: Image identifier in the Louvre system
    :param additional_metadata: Dictionary with the metadata entries to be updated
    :type additional_metadata: Dict[str, str], defaults to {}
    :param plugin_id: Plugin identifier
    :type plugin_id: str, optional
    :param e_tag: Concurrency tag
    :type e_tag: str, optional
    :param bool skip_e_tag_validation: Whether to skip concurrency tag validation
    :type skip_e_tag_validation: bool, defaults to True
    :param client_name: Client name
    :type client_name: str, defaults to python-client
    
    """

    image_id: str
    additional_metadata: Dict[str, str] = field(default_factory=dict)
    plugin_id: Optional[str] = None
    e_tag: Optional[str] = None
    skip_e_tag_validation: bool = True
    client_name: str = 'python-client'


@dataclass
class SecretPathSet:
    """
    Store secret paths
    
    :param str LOUVRE_DOMAIN: Secret path for the Louvre domain
    :param str TOKEN_ENDPOINT: Secret path for the ElvID token endpoint
    :param str CLIENT_ID_API: Secret path for the client id 
    :param str CLIENT_SECRET_API: Secret path for the client secret
    """
    LOUVRE_DOMAIN: str
    TOKEN_ENDPOINT: str
    CLIENT_ID_API: str
    CLIENT_SECRET_API: str


@dataclass
class SecretPathSets:
    """
    Store sets of secret paths

    :param SecretPathSet secret_path_set: Secret paths to be used
    :param secret_path_set_prod: Secret paths against production
    :type secret_path_set_prod: SecretPathSet, optional
    """
    secret_path_set: SecretPathSet
    secret_path_set_prod: Optional[SecretPathSet] = None

    def get_louvre_domain(self, using_production_images: bool = False) -> str:
        """
        Return the louvre domain secret path.

        :param using_production_images: If the production value is desired.
        :type using_production_images: bool, defaults to False
        :raises errors.LouvreSecretError:
        """
        if not using_production_images:
            return self.secret_path_set.LOUVRE_DOMAIN
        if using_production_images and self.secret_path_set_prod:
            return self.secret_path_set_prod.LOUVRE_DOMAIN
        raise LouvreSecretError()

    def get_token_endpoint(self, using_production_images: bool = False) -> str:
        """
        Return the token endpoint secret path.

        :param using_production_images: If the production value is desired.
        :type using_production_images: bool, defaults to False
        :raises errors.LouvreSecretError:
        """        
        if not using_production_images:
            return self.secret_path_set.TOKEN_ENDPOINT
        if using_production_images and self.secret_path_set_prod:
            return self.secret_path_set_prod.TOKEN_ENDPOINT
        raise LouvreSecretError()

    def get_client_id_api(self, using_production_images: bool = False) -> str:
        """
        Return the client id api secret path.

        :param using_production_images: If the production value is desired.
        :type using_production_images: bool, defaults to False
        :raises errors.LouvreSecretError:
        """           
        if not using_production_images:
            return self.secret_path_set.CLIENT_ID_API
        if using_production_images and self.secret_path_set_prod:
            return self.secret_path_set_prod.CLIENT_ID_API
        raise LouvreSecretError()

    def get_client_secret_api(self,
                              using_production_images: bool = False) -> str:
        """
        Return the client secret api secret path.

        :param using_production_images: If the production value is desired.
        :type using_production_images: bool, defaults to False
        :raises errors.LouvreSecretError:
        """                                 
        if not using_production_images:
            return self.secret_path_set.CLIENT_SECRET_API
        if using_production_images and self.secret_path_set_prod:
            return self.secret_path_set_prod.CLIENT_SECRET_API
        raise LouvreSecretError()



from typing import Union

from elvia_vault import VaultClient
from .errors import LouvreVaultConnError, LouvreSecretError
from .data_models import SecretPathSets


class Config():
    """        
    Class that encapsulates configuration and a connection to a vault service.
    When in cluster it uses kubernetes auth, else requires a Github token.

    :param data_models.SecretPathSets secret_path_sets: Contains SecretPathSet objects with necessary secret paths
    :param github_token: Github token
    :type github_token: str, optional

    :raises errors.LouvreVaultConnError:
    """

    _vault: VaultClient
    _secrets: dict = {}
    _secret_path_sets: SecretPathSets

    # ImageAPI - GraphQL API to fetch images
    _image_api_relative_url: str = '/image/graphql'

    # ImageEnhanceAPI - API to update image's metadata
    _image_enhance_api_relative_url: str = '/imageenhance/Metadata'

    # Token buffer in seconds
    token_buffer_seconds: int = 60

    def __init__(self,
                 secret_path_sets: SecretPathSets,
                 github_token: Union[str, None] = None):

        # Clean up, to avoid mixing secrets from different vaults, in the case
        # VAULT_ADDR changes between calls
        self._secrets = {}
        self._secret_path_sets = secret_path_sets
        # Get a fresh VaultClient instance and connect to the vault service
        try:
            self._vault = VaultClient(github_token=github_token)
        except Exception as exception:
            raise LouvreVaultConnError(str(exception))

    def get_secret(self, secret_path: str) -> str:
        """
        Return a secret value given a secret path.
        Fetch from vault only if not previously fetched.

        :param secret_path: Secret Path in the vault
        :type secret_path: str
        
        :rtype: str

        :raises errors.LouvreSecretError:
        """
        try:
            if not secret_path in self._secrets.keys():
                self._secrets[secret_path] = self._vault.get_value(secret_path)
            return self._secrets[secret_path]

        except KeyError as exception:
            raise LouvreSecretError(exception)

    def secret_path_sets(self) -> SecretPathSets:
        """
        Return the SecretPathSets object with which the client was initialised.
        
        :rtype: SecretPathSets
        :raises errors.LouvreSecretError:
        """
        if '_secret_path_sets' in vars(self):
            return self._secret_path_sets
        raise LouvreSecretError()



from datetime import datetime, timedelta
from .errors import LouvreSecretError
from typing import Dict
import requests
from .config import Config


class Token():
    """Class for working with jwt tokens."""

    _tokens: Dict[str, str] = {}
    _expiry: Dict[str, datetime] = {}

    @classmethod
    def get_token(cls,
                  config: Config,
                  using_production_images: bool = False) -> str:
        """
        Return the current token if still valid, else a new one.
        
        :param Config config: Config instance
        :param using_production_images: Whether to search in production.
        :type using_production_images: bool, defaults to False  

        :rtype: str
        :raises errors.LouvreSecretError:
        """
        endpoint: str = config.secret_path_sets().get_token_endpoint(
            using_production_images)
        if not endpoint:
            raise LouvreSecretError()
        # Assume that the token life has ended when time is not smaller than the expiry time 
        # minus the buffer
        if endpoint in cls._tokens.keys() and endpoint in cls._expiry.keys(
        ) and datetime.now() < cls._expiry[endpoint] - timedelta(
                seconds=config.token_buffer_seconds):
            return cls._tokens[endpoint]
        return cls.get_new_token(
            config=config, using_production_images=using_production_images)

    @classmethod
    def get_new_token(cls,
                      config: Config,
                      using_production_images: bool = False) -> str:
        """
        Create and get token for this API to communicate with ImageAPI and ImageEnhanceAPI.
        Assume the expiry time is found by adding expires_in to the time right before 
        triggering the request.

        :param Config config: Config instance        
        :param using_production_images: Whether to search in production.
        :type using_production_images: bool, defaults to False  
        :rtype: str    
        :raises errors.LouvreSecretError:
        """
        tick = datetime.now()
        endpoint: str = config.secret_path_sets().get_token_endpoint(
            using_production_images)

        token_response = requests.post(
            url=config.get_secret(endpoint),
            headers={},
            data={
                'grant_type':
                'client_credentials',
                'client_id':
                config.get_secret(config.secret_path_sets().get_client_id_api(
                    using_production_images)),
                'client_secret':
                config.get_secret(
                    config.secret_path_sets().get_client_secret_api(
                        using_production_images))
            })

        cls._tokens[endpoint] = token_response.json()['access_token']
        cls._expiry[endpoint] = tick + timedelta(
            seconds=int(token_response.json()['expires_in']))
        return cls._tokens[endpoint]

    @classmethod
    def get_headers(cls,
                    config: Config,
                    using_production_images: bool = False) -> dict:
        """
        Return headers to be used in HTTP requests against ImageAPI / ImageEnhanceAPI.
        
        :param Config config: Config instance        
        :param using_production_images: Whether to search in production.
        :type using_production_images: bool, defaults to False  
        :rtype: dict    
        :raises errors.LouvreSecretError:
        """
        return {
            "Authorization":
            "Bearer {}".format(
                cls.get_token(
                    config=config,
                    using_production_images=using_production_images)),
            'Content-Type':
            'application/json'
        }




class LouvreException(Exception):
    """
    Base Louvre exception class.
    """    
    pass


class LouvreImageNotFound(LouvreException):
    pass


class LouvreKeyError(LouvreException):
    pass


class LouvreQueryError(LouvreException):
    pass


class LouvreValueError(LouvreException):
    pass


class LouvreVaultConnError(LouvreException):
    pass


class LouvreSecretError(LouvreException):
    pass


class LouvreInvalidImageVariant(LouvreException):
    pass



from http import HTTPStatus
import json
import requests

from .data_models import UpdateMetadataRequest
from .errors import LouvreQueryError
from .config import Config
from .token import Token


class ImageEnhanceAPI():
    @staticmethod
    def update_metadata(config: Config,
                        update_request: UpdateMetadataRequest) -> None:
        """
        Update image metadata. Existing metadata entries with the same keys will get overwritten.

        :param Config config: Config instance
        :param UpdateMetadataRequest update_request: Metadata update request instance
        :raises errors.LouvreQueryError:
        :raises requests.RequestException:
        """
        payload = {
            "imageID":
            update_request.image_id,
            "additionalMetadataReplaceByKey": [{
                'key': k,
                'value': v
            } for k, v in update_request.additional_metadata.items()],
            "skipValidationOfProvidedETag":
            update_request.skip_e_tag_validation,
            "clientName":
            update_request.client_name
        }

        if update_request.e_tag:
            payload["eTag"] = update_request.e_tag

        if update_request.plugin_id:
            payload["pluginId"] = update_request.plugin_id

        query = json.dumps(payload)

        response = requests.put(url=config.get_secret(
            config.secret_path_sets().get_louvre_domain(False)) +
                                config._image_enhance_api_relative_url,
                                headers=Token.get_headers(config=config),
                                data=query)

        if not response.status_code == HTTPStatus.OK:
            raise LouvreQueryError(response.content)
